//  Created on : 11 Feb 2016
//  Author : M. Coroyer
//  Company : © 2016 Exoligent
//  Description:
//      This is an example of using the FipCore library.
//      The xml stations files are built from fiplabs labpro files:
//          - '2sta_worldfip_31K25.labpro'  : 31.25 Kbps
//              -> sta0_worldfip_31K25.xml
//              -> sta1_worldfip_31K25.xml
//          - '2sta_worldfip_1M.labpro'     : 1     Mbps
//              -> sta0_worldfip_1M.xml
//              -> sta1_worldfip_1M.xml
//          - '2sta_worldfip_2M5.labpro'    : 2.5   Mbps
//              -> sta0_worldfip_2M5.xml
//              -> sta1_worldfip_2M5.xml
//          - '2sta_worldfip_5M.labpro'     : 5     Mbps
//              -> sta0_worldfip_5M.xml
//              -> sta1_worldfip_5M.xml
//
//  Station0 (sta0) Configuration
//      - Arbiter: true - 1 Scan Table
//          -> Scan Table 1:
//                  SEND_ID_DAT (0x8426)
//                  SEND_ID_DAT (0x8427)
//                  SEND_ID_DAT (0x0001)
//                  SEND_MSG (5 ms)
//                  SEND_APER (10 ms)
//                  WAIT_TIME (15 ms)
//                  NEXT_MACRO ()
//      - Periodic IDs
//          -> ID(0x0001) :
//                  Type:                               Out
//                  Length:                             120 bytes
//                  Refreshment:                        true (200ms)
//                  Aperiodic's List Transfert Request: true
//                  Message Emission Channel:           false
//                  Message Reception Channel:          true
//          -> ID(0x8426) :
//                  Type:                               Out
//                  Length:                             12 bytes
//                  Refreshment:                        true (50ms)
//                  Aperiodic's List Transfert Request: false
//                  Message Emission Channel:           true (Type: Aperiodic - Channel: 0)
//                  Message Reception Channel:          false
//          -> ID(0x8427) :
//                  Type:                               In
//                  Promptness:                         true (50ms)
//                  Message Reception Channel:          false
//                  Other details: see Station1 (sta1) - ID(0x8427) 
//      - Aperiodic IDs
//          -> ID(0x8322) :
//                  Type:                               Out
//                  Length:                             10 bytes
//                  Refreshment:                        false
//                  Aperiodic's List Transfert Request: false
//                  Message Emission Channel:           false
//                  Message Reception Channel:          false
//      - Message(s)
//          -> ID (0x8426) to ID (0x8427)
//                  Source Address:                     Segment 0 - ID 0x8426
//                  Destination Address:                Segment 0 - ID 0x8427
//                  Acknowledgment:                     true
//                  Channel Emission Attached:          Type: Aperiodic - Channel: 0 (see ID(0x8426))
//
//  Station1 (sta1) Configuration
//      - Arbiter: true - 1 Scan Table
//          -> Scan Table 1:
//                  SEND_ID_DAT (0x8426)
//                  SEND_ID_DAT (0x8427)
//                  SEND_ID_DAT (0x0001)
//                  SEND_MSG (5 ms)
//                  WAIT_TIME (5 ms)
//                  NEXT_MACRO ()
//      - Periodic IDs
//          -> ID(0x8426) :
//                  Type:                               In
//                  Promptness:                         true (50ms)
//                  Message Reception Channel:          false
//                  Other details: see Station0 (sta0) - ID(0x8426) 
//          -> ID(0x8427) :
//                  Type:                               Out
//                  Length:                             12 bytes
//                  Refreshment:                        true (200 ms)
//                  Aperiodic's List Transfert Request: false
//                  Message Emission Channel:           false
//                  Message Reception Channel:          true
//      - Aperiodic IDs
//          -> ID(0x8322) :
//                  Type:                               In
//                  Promptness:                         true (50ms)
//                  Message Reception Channel:          false
//                  Other details: see Station0 (sta0) - ID(0x8322)
//
//  In this example:
//      station0_process routine:
//          Open Device / Load Configuration / Run Bus Arbiter
//          Loop
//              Each 1s 
//                  out var: ID (0x8426) value is incremented (4 first bytes / 12 total)
//                  in var:  ID (0x8427) value is read
//                  out var:  ID (0x8322) value is incremented (4 first bytes / 12 total)
//                  Events are popped (message and aperiodic var reception are treat here)
//              Each 10s 
//                  out msg: ID (0x8426) to ID (0x8427) message is sent (data: "Hello World")
//                  Aperiodic Request (ID (0x8322)) is sent
//                  Device Medium is read
//          End Loop on key press
//
//      station1_process routine:
//          Open Device / Load Configuration / Run Bus Arbiter
//          Loop
//              Each 1s 
//                  in var: ID (0x8426) value is read
//                  out var:  ID (0x8427) value is decremented (4 first bytes / 12 total)
//                  in var: ID (0x8322) value is read
//                  Events are popped (message and aperiodic var reception are treat here)
//              Each 10s 
//                  Device Medium is read
//          End Loop on key press
//
//  Note: In the example, ID(0x0001) is only used to treat the aperiodic's list requests; that's why it is
//  present in periodic window in scan table


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ExoErrors.h"
#include "FipCore.h"

#if defined(_WIN32) || defined(_WIN64)
#   include <windows.h>
#   include <conio.h>
#elif defined(__linux__) || defined(__APPLE__) || defined(__QNX__)
#   include <unistd.h>     // usleep()
#   include <termios.h>
#   include <sys/types.h>
#   include <sys/time.h>

void changemode(int dir)
{
    static struct termios oldt, newt;
    
    if (dir == 1) {
        tcgetattr(STDIN_FILENO, &oldt);
        newt = oldt;
        newt.c_lflag &= ~(ICANON | ECHO);
        tcsetattr(STDIN_FILENO, TCSANOW, &newt);
    } else
        tcsetattr(STDIN_FILENO, TCSANOW, &oldt);
}

int kbhit(void)
{
    struct timeval tv;
    fd_set rdfs;
    
    tv.tv_sec = 0;
    tv.tv_usec = 0;
    
    FD_ZERO(&rdfs);
    FD_SET(STDIN_FILENO, &rdfs);
    
    select(STDIN_FILENO + 1, &rdfs, NULL, NULL, &tv);
    return FD_ISSET(STDIN_FILENO, &rdfs);
}
#endif


uint16_t event_parameter = 0;

char* sta_xml_path[2][4] = {
    {"../xml/sta0_worldfip_31K25.xml", "../xml/sta0_worldfip_1M.xml", "../xml/sta0_worldfip_2M5.xml", "../xml/sta0_worldfip_5M.xml"},
    {"../xml/sta1_worldfip_31K25.xml", "../xml/sta1_worldfip_1M.xml", "../xml/sta1_worldfip_2M5.xml", "../xml/sta1_worldfip_5M.xml"}
};

void station0_process(uint8_t device_type, uint8_t device_index, char* xml_path);
void station1_process(uint8_t device_type, uint8_t device_index, char* xml_path);
void print_device_info(FIP_DEVICECONF* device_conf);
void print_var_conf(const USERDEF_VARIABLE* uv);
void print_conf(const FIP_USERCONF* fuc);
void event_processing();
void medium_processing();

int main(int argCount, char *argValue[])
{
    uint32_t ret = SUCCESS;
    uint8_t i = 0;
    uint8_t device_type;
    uint8_t device_index;
    uint8_t station_number;
    uint8_t station_speed;
    FIP_DEVICECONF devices_list[20];
    uint8_t nb_detected_devices = 0;
    bool    is_matched = false;
    char    xml_path[255];
    char    temp[10];

    // Device Enumeration
    printf("/*****************************************************************************/\n");
    printf("/*                          DEVICES ON COMPUTER                              */\n");
    printf("/*****************************************************************************/\n");

    ret = FC_GET_DEVICES_NUMBER(&nb_detected_devices);
    if (ret != SUCCESS) {
        printf("FC_GET_DEVICES_NUMBER - %s\n", GET_EXO_ERROR(ret));
        return 0;
    }

    printf("NUMBER OF DETECTED DEVICES - %d\n", nb_detected_devices);

    ret = FC_GET_DEVICES_LIST(devices_list, &nb_detected_devices);
    if (ret != SUCCESS) {
        printf("FC_GET_DEVICES_LIST - %s\n", GET_EXO_ERROR(ret));
        return 0;
    }

    for (i = 0; i < nb_detected_devices; i++)
        print_device_info(&(devices_list[i]));

    if (argCount < 4) {
        printf("This application must have this form: fipcore_rawframe.exe 'device_type' 'device_index' 'station_number' 'station_speed' \n");
        printf("'device_type': 1: PCI_FIP - 2: PCIE_FIP - 3: PCI104_FIP - 4: PCIE104_FIP - 5: PXI_FIP - 6: PXIE_FIP\n");
        printf("               7: PCI_OPT_FIP - 8: PCIE_OPT_FIP - 9: PCI104_OPT_FIP - 10: PCIE104_OPT_FIP - 11: PXI_OPT_FIP - 12: PXIE_OPT_FIP\n");
        printf("'device_index': 1 to 9\n");
        printf("'station_number': 0: Station 0 - 1: Station 1\n");
        printf("'station_speed': 0: 31.25 Kbps - 1: 1 Mbps - 2: 2.5 Mbps - 3: 5Mbps\n");
        return 0;
    }

    sscanf (argValue[1], "%s", temp);
    device_type = (uint8_t) atoi(temp);
    memset(temp, 0, sizeof(temp));

    sscanf (argValue[2], "%s", temp);
    device_index = (uint8_t) atoi(temp);
    memset(temp, 0, sizeof(temp));

    sscanf (argValue[3], "%s", temp);
    station_number = (uint8_t) atoi(temp);
    memset(temp, 0, sizeof(temp));

    sscanf (argValue[4], "%s", temp);
    station_speed = (uint8_t) atoi(temp);
    memset(temp, 0, sizeof(temp));
    
    if ((device_type < MIN_PCI_TYPE) || (device_type > MAX_PCI_TYPE)) {
        printf("Invalid 'device_type' parameter.\n");
        return 0;
    }
    if ((device_index < 1) || (device_index > 9)) {
        printf("Invalid 'device_index' parameter.\n");
        return 0;
    }
    if (station_number > 1) {
        printf("Invalid 'station_number' parameter.\n");
        return 0; 
    }
    if (station_speed > 3) {
        printf("Invalid 'station_speed' parameter.\n");
        return 0; 
    }

    for (i = 0; i < nb_detected_devices; i++) {

        if ((devices_list[i].device_type == device_type) && 
            (devices_list[i].device_index == device_index)) {

                if (devices_list[i].fip_speed != station_speed) {

                    if (!((devices_list[i].fip_speed == SPEED_2M5) && 
                        (station_speed == SPEED_5M))) {
                        // Support 2.5 Mbps device with 5 Mbps configuration

                        // else exit
                        printf("Station speed must match with device speed.\n");
                        return 0;
                    }
                }
                is_matched = true;
        }
    }

    if (is_matched) {

        if (station_number == 0)
            station0_process(device_type, device_index, sta_xml_path[station_number][station_speed]);
        else
            station1_process(device_type, device_index, sta_xml_path[station_number][station_speed]);
    } else
        printf("No device was found with device_type (%d) and device_index (%d).\n", device_type, device_index);

    return 0;
}

void station0_process(uint8_t device_type, uint8_t device_index, char* xml_path)
{
    uint32_t ret = SUCCESS;
    FIP_USERCONF      user_conf = {0};
    USERDEF_VARIABLE  var_conf = {0};
    PRODUCED_VARIABLE p_var = {0};
    CONSUMED_VARIABLE c_var = {0};
    MESSAGE_FRAME_OUT p_msg = {0};
    uint8_t i, j = 0;
    uint16_t src_id;
    uint16_t dst_id;
    uint16_t msg_length;
    char msg_data[] = "Hello World\n";
    uint32_t inc_data = 0;
    uint8_t loop_counter = 1;

    printf("/*****************************************************************************/\n");
    printf("/*                          STATION 0 TEST                                   */\n");
    printf("/*****************************************************************************/\n");

    ret = FC_DEVICE_OPEN(device_type, device_index);
    if (ret != SUCCESS) {
        printf("STATION 0 - %s\n", GET_EXO_ERROR(ret));
        goto _exit1;
    }

    ret = FC_RUN(XML, xml_path);
    if (ret != SUCCESS)
        goto _exit;

    ret = FC_RUN_BA(1);
    if (ret != SUCCESS)
        goto _exit;

    // Optional Configuration Information - (To execute after FC_RUN function)
    //FC_GET_CONF(&user_conf);
    //print_conf(&user_conf);
    //FC_GET_VAR_CONF(0x8426, &var_conf);
    //print_var_conf(&var_conf);

    // Output variables init
    FC_WRITE_VAR(0x0001, &p_var);
    FC_WRITE_VAR(0x8426, &p_var);
    FC_WRITE_VAR(0x8322, &p_var);

    // Output message init
    src_id = 0x8426;
    dst_id = 0x8427;

    msg_length = MSG_LENGTH_HEADER + sizeof(msg_data);
    p_msg.Msg_Mode_PF = (uint8_t) (MSG_SDN >> 8);
    p_msg.Msg_Mode_Pf = (uint8_t) MSG_SDN;
    p_msg.Msg_Length_PF = (uint8_t) (msg_length >> 8);
    p_msg.Msg_Length_Pf = (uint8_t) msg_length;
    //      Message header
    p_msg.Dest_Add_PF = (uint8_t) (dst_id >> 8);
    p_msg.Dest_Add_Pf = (uint8_t) dst_id;
    p_msg.Dest_Segment = 0;
    p_msg.Src_Add_PF = (uint8_t) (src_id >> 8);
    p_msg.Src_Add_Pf = (uint8_t) src_id;
    p_msg.Src_Segment = 0;
    for (i = 0; i < sizeof(msg_data); i++)
        p_msg.Useful_Data[i] = (uint8_t) msg_data[i];

#if defined(__linux__) || defined(__APPLE__)
    changemode(1);
#endif
    
    while (1) {

        printf("SECTION LOOP - EACH 1 SECOND\n");

        bool is_prompt = false;
        bool is_refreshed = false;

        // ID (0x8426) value is incremented (4 first bytes / 12 total)
        printf("  FC_WRITE_VAR - ID(0x8426) - DATA: ");
        memset(&p_var, 0, sizeof(PRODUCED_VARIABLE));
        *((uint32_t*) (p_var.Data_Var)) = inc_data;
        for (i = 0; i < 12; i++)
            printf(" 0x%02x", p_var.Data_Var[i]);
        printf("\n");
        
        ret = FC_WRITE_VAR(0x8426, &p_var);
        if (ret != SUCCESS)
            goto _exit;

        printf("  FC_READ_VAR  - ID(0x8427) - ");
        memset(&c_var, 0, sizeof(CONSUMED_VARIABLE));
        ret = FC_READ_VAR(0x8427, &c_var, &is_prompt, &is_refreshed);
        if (ret == SUCCESS) {
            printf("DATA: ");
            for (i = 0; i < c_var.Length; i++)
                printf(" 0x%02x", c_var.Data_Var[i]);
            printf("\n");
            printf("      [MORE INFO: IsPrompt: %d - IsRefreshed: %d - PDU: 0x%02x - Length: 0x%02x - ProductionStatus: 0x%02x]\n",
                  is_prompt, is_refreshed, c_var.Pdu, c_var.Length, c_var.Production_Status);
        } else
            printf("ERROR - %s\n", GET_EXO_ERROR(ret));

        // ID (0x8322) value is incremented (4 first bytes / 12 total)
        // Note: ID (0x8322) is written each 1s. However, production on the network will be effective
        // only each 10s (see FC_SEND_APER routine below).
        // This is due to the fact that ID (0x8322) is not present in the scan tables of the arbiter.
        // So it's an apriodic variable.
        printf("  FC_WRITE_VAR - ID(0x8322) - DATA: ");
        memset(&p_var, 0, sizeof(PRODUCED_VARIABLE));
        *((uint32_t*) (p_var.Data_Var)) = inc_data;
        for (i = 0; i < 12; i++)
            printf(" 0x%02x", p_var.Data_Var[i]);
        printf("\n");

        ret = FC_WRITE_VAR(0x8322, &p_var);
        if (ret != SUCCESS)
            goto _exit;

        inc_data++;

        event_processing(); 

        if (loop_counter == 10) { // Each 10s

            printf("SECTION LOOP - EACH 10 SECONDS\n");

            // ID (0x8426) to ID (0x8427) message is sent (data: "Hello World")
            // Send on channel 0 (aperiodic)
            printf("  FC_WRITE_MSG - From ID(0x8426) to ID(0x8427) - CH.0 - MSG: 'Hello World'\n");
            ret = FC_WRITE_MSG(0, &p_msg); 
            if (ret != SUCCESS)
                goto _exit;

            printf("  FC_SEND_APER - ID(0x8322) - URGENT\n");
            // Aperiodic Request (ID (0x8322)) is sent
            ret = FC_SEND_APER(0x8322, 1); // 1: Urgent Mode
            if (ret != SUCCESS)
                goto _exit;

            // Device Medium is read
            medium_processing(); 
        }
        
        if (kbhit())
            break;
        
#if defined(_WIN32) || defined(_WIN64)
        Sleep(1000);
#else
        usleep(1000000);
#endif
        if (loop_counter >= 10)
            loop_counter = 0;

        loop_counter++; 
        printf("\n");
    }

#if defined(__linux__) || defined(__APPLE__)
    changemode(0);
#endif
    
    ret = FC_STOP_BA();

_exit:  
    if (ret != SUCCESS)
        printf("STATION 0 - %s\n", GET_EXO_ERROR(ret));

    ret = FC_STOP();
    if (ret != SUCCESS)
        printf("STATION 0 - %s\n", GET_EXO_ERROR(ret));
_exit1:
    ret = FC_DEVICE_CLOSE();
    if (ret != SUCCESS)
        printf("STATION 0 - %s\n", GET_EXO_ERROR(ret));
}

void station1_process(uint8_t device_type, uint8_t device_index, char* xml_path)
{
    uint32_t ret = SUCCESS;
    FIP_USERCONF      user_conf = {0};
    USERDEF_VARIABLE  var_conf = {0};
    PRODUCED_VARIABLE p_var = {0};
    CONSUMED_VARIABLE c_var = {0};
    uint8_t i, j = 0;
    uint32_t dec_data = 0xFFFFFFFF;
    uint8_t loop_counter = 1;

    printf("/*****************************************************************************/\n");
    printf("/*                          STATION 1 TEST                                   */\n");
    printf("/*****************************************************************************/\n");

    ret = FC_DEVICE_OPEN(device_type, device_index);
    if (ret != SUCCESS) {
        printf("STATION 1 - %s\n", GET_EXO_ERROR(ret));
        goto _exit1;
    }

    ret = FC_RUN(XML, xml_path);
    if (ret != SUCCESS)
        goto _exit;

    ret = FC_RUN_BA(1);
    if (ret != SUCCESS)
        goto _exit;

    // Optional Configuration Information - (To execute after FC_RUN function)
    //FC_GET_CONF(&user_conf);
    //print_conf(&user_conf);
    //FC_GET_VAR_CONF(0x8426, &var_conf);
    //print_var_conf(&var_conf);

    // Output variables init
    FC_WRITE_VAR(0x8427, &p_var);

#if defined(__linux__) || defined(__APPLE__)
    changemode(1);
#endif
    
    while (1) {

        printf("SECTION LOOP - EACH 1 SECOND\n");

        bool is_prompt = false;
        bool is_refreshed = false;

        printf("  FC_READ_VAR  - ID(0x8426) - ");
        memset(&c_var, 0, sizeof(CONSUMED_VARIABLE)); 
        ret = FC_READ_VAR(0x8426, &c_var, &is_prompt, &is_refreshed);
        if (ret == SUCCESS) {
            printf("DATA: ");
            for (i = 0; i < c_var.Length; i++)
                printf(" 0x%02x", c_var.Data_Var[i]);
            printf("\n");
            printf("      [MORE INFO: IsPrompt: %d - IsRefreshed: %d - PDU: 0x%02x - Length: 0x%02x - ProductionStatus: 0x%02x]\n",
                  is_prompt, is_refreshed, c_var.Pdu, c_var.Length, c_var.Production_Status);
        } else
            printf("ERROR - %s\n", GET_EXO_ERROR(ret));

        // ID (0x8427) value is decremented (4 first bytes / 12 total)
        printf("  FC_WRITE_VAR - ID(0x8427) - DATA: ");
        memset(&p_var, 0, sizeof(PRODUCED_VARIABLE));
        *((uint32_t*) (p_var.Data_Var)) = dec_data;
        for (i = 0; i < 12; i++)
            printf(" 0x%02x", p_var.Data_Var[i]);
        printf("\n");
        
        
        ret = FC_WRITE_VAR(0x8427, &p_var);
        if (ret != SUCCESS)
            goto _exit;

        printf("  FC_READ_VAR  - ID(0x8322) - ");
        memset(&c_var, 0, sizeof(CONSUMED_VARIABLE));
        ret = FC_READ_VAR(0x8322, &c_var, &is_prompt, &is_refreshed);
        if (ret == SUCCESS) {
            printf("DATA: ");
            for (i = 0; i < c_var.Length; i++)
                printf(" 0x%02x", c_var.Data_Var[i]);
            printf("\n");
            printf("      [MORE INFO: IsPrompt: %d - IsRefreshed: %d - PDU: 0x%02x - Length: 0x%02x - ProductionStatus: 0x%02x]\n",
                  is_prompt, is_refreshed, c_var.Pdu, c_var.Length, c_var.Production_Status);
        } else
            printf("ERROR - %s\n", GET_EXO_ERROR(ret));

        dec_data--;

        event_processing(); 

        if (loop_counter == 10) { // Each 10s

            printf("SECTION LOOP - EACH 10 SECONDS\n");

            // Device Medium is read
            medium_processing(); 
        }
        
        if (kbhit())
            break;
        
#if defined(_WIN32) || defined(_WIN64)
        Sleep(1000);
#else
        usleep(1000000);
#endif
        if (loop_counter >= 10)
            loop_counter = 0;

        loop_counter++;
        printf("\n");
    }

#if defined(__linux__) || defined(__APPLE__)
    changemode(0);
#endif
    
    ret = FC_STOP_BA();

_exit:  
    if (ret != SUCCESS)
        printf("STATION 1 - %s\n", GET_EXO_ERROR(ret));

    ret = FC_STOP();
    if (ret != SUCCESS)
        printf("STATION 1 - %s\n", GET_EXO_ERROR(ret));
_exit1:
    ret = FC_DEVICE_CLOSE();
    if (ret != SUCCESS)
        printf("STATION 1 - %s\n", GET_EXO_ERROR(ret));
}

void event_processing()
{
    uint32_t ret = SUCCESS;
    uint16_t event_code = 0;
    
    ret = FC_READ_EVT(&event_code, &event_parameter);

//      event_code:
//      EVT_SEND_VAR_P      0x8100  /* Sending of a variable - Permanent */
//      EVT_SEND_VAR_T      0x0100  /* Sending of a variable - Temporary */
//      EVT_RECEIVE_VAR_P   0x8200  /* Reception of a variable - Permanent */
//      EVT_RECEIVE_VAR_T   0x0200  /* Reception of a variable - Temporary */
//      EVT_RECEIVE_MSG     0x0240  /* Reception of a message */
//      EVT_SEND_MSG        0x0140  /* Sending of a message */
//      EVT_SEND_APU        0x0130  /* Sending of a list of urgent aperiodic requests */
//      EVT_SEND_APN        0x0131  /* Sending of a list of normal aperiodic requests */
//      EVT_BA_ACTIVITY     0x0400  /* BA reactivation (active on network) */
//      EVT_BA_STOP1        0x0401  /* Stopping of BA following timeout */
//      EVT_BA_STOP2        0x0402  /* Stopping of BA following network anomalies */
//      EVT_BA_STOP3        0x0404  /* Stopping of BA following user request */
//      EVT_BA_IDLE         0x0408  /* Return of BA to idle status */
//      EVT_BA_SUSPEND0     0x0500  /* Suspension of BA program following SUSPEND number 1 */
//      EVT_BA_SUSPEND1     0x0501  /* Suspension of BA program following SUSPEND number 2 */
//      EVT_BA_SUSPEND2     0x0502  /* Suspension of BA program following SUSPEND number 3 */
//      ...    
//      EVT_BA_SUSPEND255   0x05ff  /* Suspension of BA program following SUSPEND number 256 */

    while (ret == SUCCESS) { // All events are popped

        bool msgReceivedToTreat = false;

        printf("  FC_READ_EVT - ");
        switch (event_code) {
        case EVT_SEND_VAR_P:    printf("SEND_VAR_P - AccessKey: 0x%04x\n", event_parameter); break;
        case EVT_SEND_VAR_T:    printf("SEND_VAR_T - AccessKey: 0x%04x\n", event_parameter); break;
        case EVT_RECEIVE_VAR_P: printf("RECEIVE_VAR_P\n"); break;
        case EVT_RECEIVE_VAR_T: printf("RECEIVE_VAR_T\n"); break;
        case EVT_RECEIVE_MSG:   printf("RECEIVE_MSG\n"); msgReceivedToTreat = true; break;
        case EVT_SEND_MSG:      printf("SEND_MSG - Report: 0x%04x\n", event_parameter); break;
        case EVT_SEND_APU:      printf("SEND_APU\n"); break;
        case EVT_SEND_APN:      printf("SEND_APN\n"); break;
        case EVT_BA_ACTIVITY:   printf("BA_ACTIVITY\n"); break;
        case EVT_BA_STOP1:      printf("BA_STOP1\n"); break;
        case EVT_BA_STOP2:      printf("BA_STOP2\n"); break;
        case EVT_BA_STOP3:      printf("BA_STOP3\n"); break;
        case EVT_BA_IDLE:       printf("BA_IDLE\n"); break;
        default:
            if ((event_code & 0x0500) == 0x0500) {
                uint8_t suspend_number = (uint8_t) (event_code & 0x00FF);
                printf("BA_SUSPEND%d - BA Address Program: 0x%04x\n", suspend_number, event_parameter);
            } else
                printf("UNKNOWN - EventCode: 0x%04x - EventParameter: 0x%04x\n", event_code, event_parameter);          

        break;
        }

        if (msgReceivedToTreat) {
            uint32_t ret1 = SUCCESS;
            uint16_t i = 0;
            MESSAGE_FRAME_IN c_msg = {0};

            ret = FC_READ_MSG(&c_msg); 
            if (ret == SUCCESS) {

                uint8_t srcSeg  = c_msg.Src_Segment;
                uint8_t destSeg = c_msg.Dest_Segment;
                uint16_t srcId  = (((uint16_t) c_msg.Src_Add_PF) << 8)  + c_msg.Src_Add_Pf;
                uint16_t destId = (((uint16_t) c_msg.Dest_Add_PF) << 8) + c_msg.Dest_Add_Pf;

                printf("  FC_READ_MSG - From ID(0x%04x) to ID(0x%04x) - ", srcId, destId);

                printf("MSG: ");
                for (i = 0; i < c_msg.Msg_Length; i++)
                    printf("%c", c_msg.Useful_Data[i]);
                printf("\n");
            } else
                printf("ERROR - %s\n", GET_EXO_ERROR(ret));
        }

        ret = FC_READ_EVT(&event_code, &event_parameter);
    }
}

void medium_processing()
{
    uint32_t ret = SUCCESS;
    FIP_NETWORK_STATUS  fipNetworkStatus = {0};

    // Lecture du medium
    ret = FC_GET_MEDIUM(&fipNetworkStatus);
    if (ret != SUCCESS)
        printf("  FC_GET_MEDIUM error: %08x", ret);
    else {
        bool ch1Active = ((fipNetworkStatus.FIELDUAL_Status & MSK_CHANNEL_STATUS_CHANNEL1_VALID) == MSK_CHANNEL_STATUS_CHANNEL1_VALID);
        bool ch2Active = ((fipNetworkStatus.FIELDUAL_Status & MSK_CHANNEL_STATUS_CHANNEL2_VALID) == MSK_CHANNEL_STATUS_CHANNEL2_VALID);
        bool ch1TXError = ((fipNetworkStatus.FIELDUAL_Status & MSK_CHANNEL_STATUS_CHANNEL1_TX_ERROR) == MSK_CHANNEL_STATUS_CHANNEL1_TX_ERROR);
        bool ch2TXError = ((fipNetworkStatus.FIELDUAL_Status & MSK_CHANNEL_STATUS_CHANNEL2_TX_ERROR) == MSK_CHANNEL_STATUS_CHANNEL2_TX_ERROR);
        bool ch1WatchdogError = ((fipNetworkStatus.FIELDUAL_Status & MSK_CHANNEL_STATUS_CHANNEL1_WATCHDOG) == MSK_CHANNEL_STATUS_CHANNEL1_WATCHDOG);
        bool ch2WatchdogError = ((fipNetworkStatus.FIELDUAL_Status & MSK_CHANNEL_STATUS_CHANNEL2_WATCHDOG) == MSK_CHANNEL_STATUS_CHANNEL2_WATCHDOG);
        char baStatusStr[255];

        switch (fipNetworkStatus.BA_Status) {
            case BA_STOPPED:        strcpy(baStatusStr, "STOPPED"); break;
            case BA_STARTING:       strcpy(baStatusStr, "STARTING"); break;
            case BA_IDLE:           strcpy(baStatusStr, "IDLE"); break;
            case BA_MACRO_END:      strcpy(baStatusStr, "MACRO END"); break;
            case BA_SENDING_ID_DAT: strcpy(baStatusStr, "SENDING ID_DAT"); break;
            case BA_SENDING_ID_MSG: strcpy(baStatusStr, "SENDING ID_MSG"); break;
            case BA_TESTING:        strcpy(baStatusStr, "TESTING"); break;
            case BA_TESTING_END:    strcpy(baStatusStr, "TESTING END"); break;
            case BA_PENDING:        strcpy(baStatusStr, "PENDING"); break;
            case BA_APER_WINDOW:    strcpy(baStatusStr, "APER. VAR WINDOW"); break;
            case BA_MSG_WINDOW:     strcpy(baStatusStr, "APER. MSG WINDOW"); break;
            case BA_WAITING_TIME:   strcpy(baStatusStr, "WAITING TIME"); break;
            case BA_WAITING_SYNC:   strcpy(baStatusStr, "WAITING SYNC."); break;
            case BA_WAITING_SYNC_SILENT: strcpy(baStatusStr, "WAITING SYNC. SILENT"); break;
            default: strcpy(baStatusStr, "UNKNOWN"); break;

        }

        printf("  FC_GET_MEDIUM\n");
        printf("    No_Error_Channel1   : %d\n", fipNetworkStatus.No_Error_Channel1);
        printf("    No_Error_Channel2   : %d\n", fipNetworkStatus.No_Error_Channel2);
        printf("    Error_Channel1      : %d\n", fipNetworkStatus.Error_Channel1);
        printf("    Error_Channel2      : %d\n", fipNetworkStatus.Error_Channel2);
        printf("    Error_No_Echo       : %d\n", fipNetworkStatus.Error_No_Echo);
        printf("    Error_Testp_Channel1: %d\n", fipNetworkStatus.Error_Testp_Channel1);
        printf("    Error_Testp_Channel2: %d\n", fipNetworkStatus.Error_Testp_Channel2);
        printf("    BA_Current_Macrocyle: %04xh\n", fipNetworkStatus.BA_Current_Macrocyle);
        printf("    BA_Status           : %04xh - [%s]\n", fipNetworkStatus.BA_Status, baStatusStr);
        printf("    FIELDUAL_Status     : %04xh\n", fipNetworkStatus.FIELDUAL_Status);
        printf("          [Channel 1 - Active: %d - TX Error: %d - Watchdog Error: %d]\n", ch1Active, ch1TXError, ch1WatchdogError);
        printf("          [Channel 2 - Active: %d - TX Error: %d - Watchdog Error: %d]\n", ch2Active, ch2TXError, ch2WatchdogError);


    }
}

void print_device_info(FIP_DEVICECONF* device_conf)
{
    switch(device_conf->device_type) {
    case PCI_FIP:           printf("   device_type     : PCI_FIP \n"); break;
    case PCIE_FIP:          printf("   device_type     : PCIE_FIP \n"); break;
    case PCI104_FIP:        printf("   device_type     : PCI104_FIP \n"); break;
    case PCIE104_FIP:       printf("   device_type     : PCIE104_FIP \n"); break;
    case PXI_FIP:           printf("   device_type     : PXI_FIP \n"); break;
    case PXIE_FIP:          printf("   device_type     : PXIE_FIP \n"); break;
    case PCI_OPT_FIP:       printf("   device_type     : PCI_OPT_FIP \n"); break;
    case PCIE_OPT_FIP:      printf("   device_type     : PCIE_OPT_FIP \n"); break;
    case PCI104_OPT_FIP:    printf("   device_type     : PCI104_OPT_FIP \n"); break;
    case PCIE104_OPT_FIP:   printf("   device_type     : PCIE104_OPT_FIP \n"); break;
    case PXI_OPT_FIP:       printf("   device_type     : PXI_OPT_FIP \n"); break;
    case PXIE_OPT_FIP:      printf("   device_type     : PXIE_OPT_FIP \n"); break;
    case MOBI_FIP:          printf("   device_type     : MOBI_FIP \n"); break;
    default:                printf("   device_type     : UNKNOWN_TYPE \n"); break;
    }
    printf("   device_index    : 0x%04x \n", device_conf->device_index);
    printf("   product_number  : 0x%08x \n", device_conf->product_number);
    printf("   device_version  : %d.%d.%d \n", device_conf->device_version.major,
                                            device_conf->device_version.minor,
                                            device_conf->device_version.revision);
    switch(device_conf->fip_speed) {
    case SPEED_31K25:   printf("   fip_speed       : SPEED_31K25 \n"); break;
    case SPEED_1M:      printf("   fip_speed       : SPEED_1M \n"); break;
    case SPEED_2M5:     printf("   fip_speed       : SPEED_2M5 or SPEED_5M \n"); break;
    default:            printf("   fip_speed       : UNKNOWN_SPEED \n"); break;
    }
    switch(device_conf->fip_impedance) {
    case IMPEDANCE_150: printf("   fip_impedance   : IMPEDANCE_150 \n\n"); break;
    case IMPEDANCE_120: printf("   fip_impedance   : IMPEDANCE_120 \n\n"); break;
    default:            printf("   fip_impedance   : UNKNOWN_IMPEDANCE \n\n"); break;
    }
}

void print_var_conf(const USERDEF_VARIABLE* uv) {
    if (uv != NULL) {
        printf("   identifier         : 0x%04x \n", uv->identifier);
        printf("   pdu                : 0x%04x \n", uv->pdu);
        printf("   length             : %d \n",     uv->length);
        printf("   type               : %d \n",     uv->type);
        printf("   refreshment_type   : %d \n",     uv->refreshment_type);
        printf("   refreshment_period : 0x%04x \n", uv->refreshment_period);
        printf("   promptitude_type   : %d \n",     uv->promptitude_type);
        printf("   promptitude_period : 0x%04x \n", uv->promptitude_period);
        printf("   msg_channel        : %d \n",     uv->msg_channel);
        printf("   messages_service   : 0x%02x \n\n", uv->messages_service);
    } else {
        printf("    - user_def_var    : \n");
        printf("        - unallocated \n");
    }
}

void print_conf(const FIP_USERCONF* fuc)
{
    uint16_t cp = 0;
    uint16_t i  = 0;


    printf("/*****************************************************************************/\n");
    printf("/*                              FIP_USERCONF                                 */\n");
    printf("/*****************************************************************************/\n");
    if (fuc != NULL) {
        printf("    - this_station               : \n");
        printf("        - subscriber_number      : %d \n", fuc->this_station->subscriber_number);
        printf("        - subscriber_name        : %s \n", fuc->this_station->subscriber_name);
        printf("    - speed                      : %d \n", fuc->speed);
        printf("    - t_slot                     : %d \n", fuc->t_slot);
        printf("    - silence                    : %d \n", fuc->silence);
        printf("    - turn_around                : %d \n", fuc->turn_around);
        printf("    - frame_type                 : %d \n", fuc->frame_type);
        printf("    - is_arbiter                 : %d \n", fuc->is_arbiter);
        printf("    - arbiter                    : \n");
        printf("        - ba_max_subscriber_number : %d \n", fuc->arbiter->ba_max_subscriber_number);
        printf("        - ba_priority            : %d \n", fuc->arbiter->ba_priority);
        printf("        - ba_max_priority        : %d \n", fuc->arbiter->ba_max_priority);
        printf("        - ba_optimization        : %d \n", fuc->arbiter->ba_optimization);
        printf("        - nb_ba_1_instructions   : %d \n", cp = fuc->arbiter->nb_ba_1_instructions);
        printf("        - ba_1_instructions      : \n");
        {
            INSTRUCTION* instructions = fuc->arbiter->ba_1_instructions;
            for (i = 0; i < cp; i++)
                printf("            - %s \n", (instructions + i)->op_code);
        }
        printf("        - nb_ba_2_instructions   : %d \n", cp = fuc->arbiter->nb_ba_2_instructions);
        printf("        - ba_2_instructions      : \n");
        {
            INSTRUCTION* instructions = fuc->arbiter->ba_2_instructions;
            for (i = 0; i < cp; i++)
                printf("            - %s \n", (instructions + i)->op_code);
        }   
        printf("    - nb_variables               : %d \n", cp = fuc->nb_variables);
        printf("    - variables                  : \n");
        {
            for (i = 0; i < cp; i++) {
                printf("        - number %d \n", i);
                printf("            - identifier         : 0x%04x \n", fuc->variables[i].identifier);
                printf("            - pdu                : 0x%04x \n", fuc->variables[i].pdu);
                printf("            - length             : %d \n", fuc->variables[i].length);
                printf("            - type               : %d \n", fuc->variables[i].type);
                printf("            - refreshment_type   : %d \n", fuc->variables[i].refreshment_type);
                printf("            - refreshment_period : 0x%04x \n", fuc->variables[i].refreshment_period);
                printf("            - promptitude_type   : %d \n", fuc->variables[i].promptitude_type);
                printf("            - promptitude_period : 0x%04x \n", fuc->variables[i].promptitude_period);
                printf("            - msg_channel        : %d \n", fuc->variables[i].msg_channel);
                printf("            - messages_service   : 0x%02x \n", fuc->variables[i].messages_service);
            }
        }
        printf("    - nb_messages                : %d \n", cp = fuc->nb_messages);
        printf("    - messages                   : \n");
        {
            for (i = 0; i < cp; i++) {
                printf("        - number %d \n", i);
                printf("            - type               : %d \n", fuc->messages[i].type);
                printf("            - channel            : %d \n", fuc->messages[i].channel);
                printf("            - acknowledgment     : %d \n", fuc->messages[i].acknowledgment);
                printf("            - destination_identifier: 0x%04x  \n", fuc->messages[i].destination_identifier);
                printf("            - destination_segment: 0x%02x  \n", fuc->messages[i].destination_segment);
                printf("            - source_identifier  : 0x%04x  \n", fuc->messages[i].source_identifier);
                printf("            - source_segment     : 0x%02x  \n", fuc->messages[i].source_segment);
            }
        }
        printf("     - nb_tags            : %d \n", cp = fuc->nb_tags);
        printf("     - tags               : \n");
        {
            for (i = 0; i < cp; i++) {
                printf("        - number %d \n", i);
                printf("            - tag_name      : %s \n", fuc->tags[i].tag_name);
                printf("            - tag_id        : %d \n", fuc->tags[i].tag_id);
                printf("            - type_name     : %s \n", fuc->tags[i].type_name);
                printf("            - type_id       : %d \n", fuc->tags[i].type_id);
                printf("            - access_rights : %d \n", fuc->tags[i].access_rights);
                printf("            - description   : %s \n", fuc->tags[i].description);
                printf("            - identifier    : 0x%04x \n", fuc->tags[i].identifier);
                printf("            - start_bit     : %d \n", fuc->tags[i].start_bit);
                printf("            - stop_bit      : %d \n", fuc->tags[i].stop_bit);
                printf("            - endianness    : %d \n", fuc->tags[i].endianness);
            }
        }
    } else {
        printf("    - user_conf               : \n");
        printf("        - unallocated \n");
    }
}