// (C) Copyright Exoligent 2017.
// Authors and Contributors: Maxime Coroyer
// -- Common structures to FipCodeAccess / FipAtomAccess ...
#ifndef _FIPACCESSCOMMON_H_
#define _FIPACCESSCOMMON_H_

#include <stdint.h>
#include <stdbool.h>
#include "EnumCommon.h"
#include "FipCodeConstants.h"

#if defined(_WIN32) || defined(_WIN64)
#   include <windows.h>
    typedef HANDLE  device_handle_t;
#else // defined(__linux__) || defined(__APPLE__)
    typedef int     device_handle_t;
#endif

typedef uint32_t (*cb_completion)(uint8_t direction, uint16_t dest_id, uint8_t dest_seg, uint16_t src_id, uint8_t src_seg,
                                  uint32_t error, double transaction_duration_s, char* target_path, void* callback_ctx);
typedef void (*cb_progression)( uint8_t direction, uint16_t dest_id, uint8_t dest_seg, uint16_t src_id, uint8_t src_seg,
                                uint32_t chunk_number, uint32_t total_chunks, double time_elapsed_s, void* callback_ctx);

#pragma pack(1) // 1 byte alignment

//***************************************************************************************
//
//                              CONFIGURATION STRUCTURES
//
//***************************************************************************************
typedef struct
{
    char op_code[20];

} INSTRUCTION;

typedef struct
{
    char        tag_name[MAX_NAME_LENGTH];      // Tag name
    uint16_t    tag_id;                         // Tag ID
    char        type_name[MAX_TYPE_LENGTH];     // Tag type name (ex: Bool, Int8, UInt8 ...)
    uint16_t    type_id;                        // Tag type ID (ex: TAG_TYPES_BOOLEAN)
    uint8_t     access_rights;                  // Access Rights: 0: Read/Writeable, 1: Readable
    char        description[MAX_UNIT_LENGTH];   // Tag description: value unit, comment ...
    uint16_t    identifier;                     // FIP variable identifier
    uint16_t    start_bit;                      // Start bit value (FIP frame)
    uint16_t    stop_bit;                       // Stop bit value (FIP frame)
    int8_t      endianness;                     // Endianness: -1: not significant, 0: big-endian, 1: little-endian
} USERDEF_TAG;

typedef struct
{
    uint16_t    identifier;                     // FIP variable identifier
    uint16_t    pdu;                            // PDU type: (0x40: MPS PDU, 0x50: SM MPS PDU)
    uint16_t    length;                         // Length in bytes
    uint8_t     type;                           // 0: produced, 1: consumed
    uint8_t     refreshment_type;               // Refreshment: 0: no, 1: yes, 2: yes (dynamic)
    uint16_t    refreshment_period;             // Refreshment period
    uint8_t     promptitude_type;               // Promptness: 0: no, 1: yes
    uint16_t    promptitude_period;             // Promptness period
    int16_t     msg_channel;                    // Message channel: -1: no channel, 0: aperiodic channel, 1..8: periodic channel
    uint8_t     messages_service;               // Messages service:
                                                // bit1=1   -> Messaging requests authorized
                                                // bit0=1   -> Message reception authorized
                                                // bit6=1   -> MPS aperiodic requests authorized
} USERDEF_VARIABLE; // (VARIABLE4)

typedef struct
{
    // Message characteristics
    uint8_t     type;                           // b7=1 -> Message reserved for file transaction
                                                // b7=0 -> Normal message
                                                // b0=1 -> Message to transmit
                                                // b0=0 -> Message to receive
    uint8_t     channel;                        // Message channel: 0: aperiodic channel, 1..8: periodic channel
                                                // Note: Only significant for a message to transmit)
    uint8_t     acknowledgment;                 // Acknowledgment request: 0: no, 1: yes
                                                // Note: Only significant for a message to transmit)
    // Header definition
    uint16_t    destination_identifier;
    uint8_t     destination_segment;
    uint16_t    source_identifier;
    uint8_t     source_segment;

} USERDEF_MESSAGE; // (MESSAGE4)

typedef struct
{
    uint8_t subscriber_number;
    char    subscriber_name[MAX_NAME_LENGTH];

} USERDEF_STATION_INF; // (INFO_STATION4)

typedef struct
{
    uint8_t         ba_max_subscriber_number;
    uint8_t         ba_priority;
    uint8_t         ba_max_priority;
    uint8_t         ba_optimization;
    uint16_t        nb_ba_1_instructions;
    INSTRUCTION*    ba_1_instructions;
    uint16_t        nb_ba_2_instructions;
    INSTRUCTION*    ba_2_instructions;

} USERDEF_ARBITER;

typedef struct
{
    USERDEF_STATION_INF* this_station;

    uint8_t             speed;          // = Speed - 0: 31.25 kb/s, 1: 1 Mb/s, 2: 2.5 Mb/s, 3: 5 Mb/s
    uint8_t             t_slot;         // = TSlot - 0, 1, 2 ou 3
    uint8_t             silence;        // = Silence_Par (max: 3Fh)
    uint8_t             turn_around;    // = RPTime_Par (max: 3Fh)
    uint8_t             frame_type;     // 0: FIP - 1: WorldFIP - 2: SlowFIP (FipCode5 1MHz compatibility)
    bool                is_arbiter;     // Arbiter presence: 0: 0, 1: yes
    USERDEF_ARBITER*    arbiter;        // FIP Arbitrator descriptor
    uint16_t            nb_variables;   // Number of FIP MPS variables
    USERDEF_VARIABLE*   variables;      // Descriptors of the FIP MPS variables
    uint16_t            nb_messages;    // Number of FIP messages
    USERDEF_MESSAGE*    messages;       // Descriptors of the FIP messages
    uint16_t            nb_tags;        // Number of FIP tags
    USERDEF_TAG*        tags;           // Descriptors of the FIP tags (for FIP variables)

} FIP_USERCONF; // User FIP Configuration (xml/cnf files)

typedef struct
{
    uint8_t             major;
    uint8_t             minor;
    uint8_t             revision;
} DEVICE_VERSION;

typedef struct
{
    EXO_DEVICE_TYPE     device_type;        // Device type to handle
    uint8_t             device_index;       // Device index to handle
    uint8_t             device_is_used;     // 1: Device in use
                                            // 0: Free device
    device_handle_t     device_handle;      // Handle on device driver
    uint32_t            product_number;     // Product number of the device
    DEVICE_VERSION      device_version;     // Version number of the device
    FIP_SPEED           fip_speed;          // FIP speed
    FIP_IMPEDANCE       fip_impedance;      // FIP impedance
} FIP_DEVICECONF;


//***************************************************************************************
//
//                              COMMUNICATION STRUCTURES  
//
//***************************************************************************************

//***************************************************************************************
// CONSUMED_VARIABLE
// Structure related to a consumed variable. - byte array [cf FIPCODE v6_4 p.112]
// <CONSUMED_VARIABLE> ::=      <Pdu><Length>{<Data_Var>}^Length-2
//                          or  <Pdu><Length>{<Data_Var>}^Length-3<Production_Status>
//                          if refreshment is active
// Used in: 
//  - FIP_READ_VAR
//  - FIP_READ_VAR_SVF
//  - USER_READ_VAR
//***************************************************************************************
typedef struct
{
    uint8_t Pdu;
    uint8_t Length;
    uint8_t Data_Var[NB_BYTE_PER_SEGMENT];
    uint8_t Production_Status;

} CONSUMED_VARIABLE;

//***************************************************************************************
// CONSUMED_VARIABLE_TIME
// Structure related to a consumed variable with dynamic refreshment.
// byte array [cf FIPCODE v6_4 p.118].
// <CONSUMED_VARIABLE_TIME> ::= 
//              <Pdu><Length>{<Data_Var>}^Length-7<Production_Time><Production_Status>
//
// Used in:
//  - FIP_READ_VAR_TIME
//  - USER_READ_VAR_TIME
//***************************************************************************************
typedef struct
{
    uint8_t Pdu;
    uint8_t Length;
    uint8_t Data_Var[NB_BYTE_PER_SEGMENT];
    uint32_t Production_Time;
    uint8_t Production_Status;

} CONSUMED_VARIABLE_TIME;

//***************************************************************************************
// PRODUCED_VARIABLE
// Structure related to a produced variable - byte array [cf FIPCODE v6_4 p.126]
// <PRODUCED_VARIABLE> ::=      {<Data_Var>}^n                      n = [6 ... 126]
//                          ou  {<Data_Var>}^n<Production_Status>   n = [1 ... 125]
// Note: PRODUCED_VARIABLE can be used as PRODUCED_VARIABLE_TIME. In this case,
// <Production_Time> and <Production_Status> must be inserted after useful data in Data_Var.
//
// Used in:
//  - FIP_WRITE_VAR
//  - USER_WRITE_VAR
//***************************************************************************************
typedef struct
{
    uint8_t Data_Var[NB_BYTE_PER_SEGMENT - 2];

} PRODUCED_VARIABLE;

//***************************************************************************************
// PRODUCED_VARIABLE_TIME
// Structure related to a produced variable with dynamic refreshment.
// byte array [cf FIPCODE v6_4 p.126].
// <PRODUCED_VARIABLE_TIME> ::= 
//              {<Data_Var>}^n<Production_Time><Production_Status>  n = [6 ... 120]
//
// Used in:
//  - USER_WRITE_VAR_TIME
//***************************************************************************************
typedef struct
{
    uint8_t Data_Var[NB_BYTE_PER_SEGMENT];
    uint8_t Production_Time;
    uint8_t Production_Status;

} PRODUCED_VARIABLE_TIME;

//***************************************************************************************
// PRESENT_LIST_BI_MEDIUM
// Structure containing the list of subscriber number (SM_MPS presence variable)
// for BI-MEDIUM.
// byte array [cf FIPCODE v6_4 p.122].
// Note: PRESENT_LIST is used as generic structure to contain bimedium or monomedium
// presence list.
// Example:
// Medium_Tabi[0].Bit0 = 1      -> Subscriber 0 is present on the medium i (i = 1 or 2)
// Medium_Tabi[0].Bit7 = 0      -> Subscriber 7 is absent on the medium i
// Medium_Tabi[31].Bit7 = 1     -> Subscriber 247 is present on the medium i
// Medium_Tabi[31].Bit7 = 1     -> Subscriber 255 is present on the medium i
//
// Used in:
//  - FIP_READ_LSTPRESENT
//  - USER_READ_LSTPRESENT
//***************************************************************************************
typedef struct
{
    uint8_t Pdu;                        // SM_MPS_PDU
    uint8_t Total_Length;               // 0x46
    uint8_t Medium_Tab1_Id;             // 0x80
    uint8_t Medium_Tab1_Length;         // 0x20     
    uint8_t Medium_Tab1[32];            // Presence list of the medium 1
    uint8_t Medium_Tab2_Id;             // 0x81
    uint8_t Medium_Tab2_Length;         // 0x20
    uint8_t Medium_Tab2[32];            // Presence list of the medium 2

} PRESENT_LIST_BI_MEDIUM, PRESENT_LIST;

//***************************************************************************************
// PRESENT_LIST_MONO_MEDIUM
// Structure containing the list of subscriber number (SM_MPS presence variable)
// for MONO-MEDIUM.
// byte array [cf FIPCODE v6_4 p.122].
//
// Medium_Tab[0].Bit0 = 1      -> Subscriber 0 is present on the medium
// Medium_Tab[0].Bit7 = 0      -> Subscriber 7 is absent on the medium
// Medium_Tab[31].Bit7 = 1     -> Subscriber 247 is present on the medium
// Medium_Tab[31].Bit7 = 1     -> Subscriber 255 is present on the medium
//
// Used in:
//  - FIP_READ_LSTPRESENT
//  - USER_READ_LSTPRESENT
//***************************************************************************************
typedef struct
{
    uint8_t Pdu;                        // SM_MPS_PDU
    uint8_t Total_Length;               // 0x46
    uint8_t Medium_Tab_Id;              // 0x80
    uint8_t Medium_Tab_Length;          // 0x20     
    uint8_t Medium_Tab[32];             // Presence list of the medium

} PRESENT_LIST_MONO_MEDIUM;

//***************************************************************************************
// MESSAGE_FRAME_OUT
// Structure defining the body of a message to be sent [cf FIPCODE v6_4 p.155]
//
// Used in:
//  - FIP_WRITE_MSG
//  - USER_WRITE_MSG
//***************************************************************************************
typedef struct
{
    // Characteristics
    uint8_t Msg_Mode_PF;        // MSG_SDN ou MSG_SDA (respectively without/with acknowledgment)
    uint8_t Msg_Mode_Pf;
    uint8_t Msg_Length_PF;      // Length = Header + payload data
    uint8_t Msg_Length_Pf;
    // Message header
    uint8_t Dest_Add_PF;
    uint8_t Dest_Add_Pf;
    uint8_t Dest_Segment;
    uint8_t Src_Add_PF;
    uint8_t Src_Add_Pf;
    uint8_t Src_Segment;
    // Payload data
    uint8_t Useful_Data[MAX_MSG_LENGTH_WITHOUT_CHARACT_AND_HEADER];

} MESSAGE_FRAME_OUT;

//***************************************************************************************
// MESSAGE_FRAME_IN
// Structure defining the body of a received message [cf FIPCODE v6_4 p.161]
//
// Used in:
//  - FIP_READ_MSG
//  - USER_READ_MSG
//***************************************************************************************
typedef struct
{
    // Caractéristiques du message
    uint16_t Msg_Reception_Time;
    uint16_t Msg_Length;
    // En-tête du message
    uint8_t Dest_Add_PF;
    uint8_t Dest_Add_Pf;
    uint8_t Dest_Segment;
    uint8_t Src_Add_PF;
    uint8_t Src_Add_Pf;
    uint8_t Src_Segment;
    // Données utiles
    uint8_t Useful_Data[MAX_MSG_LENGTH_WITHOUT_CHARACT_AND_HEADER];

} MESSAGE_FRAME_IN;

//***************************************************************************************
// BA_PARAMETERS
// Structure of general parameters of the arbiter "on-line" editable
// [cf FIPCODE v6_4 p.192].
//
// Used in:
//  - FIP_CHANGE_PARAM_BA
//  - USER_CHANGE_PARAM_BA
//***************************************************************************************
typedef struct
{
    uint16_t BA_Subscriber_Number;
    uint16_t BA_Max_Subscriber_Number;
    uint16_t BA_Priority;
    uint16_t BA_Start_Time;
    uint16_t BA_Election_Time;

} BA_PARAMETERS;

//***************************************************************************************
// FIP_NETWORK_STATUS
// FIP Network Status Table - 10-word [cf FIPCODE v6_4 p.81]
//***************************************************************************************
typedef struct
{
    // -- Channel Counters
    uint16_t No_Error_Channel1;
    uint16_t No_Error_Channel2;
    uint16_t Error_Channel1;
    uint16_t Error_Channel2;
    uint16_t Error_No_Echo;
    uint16_t Error_Testp_Channel1;
    uint16_t Error_Testp_Channel2;
    // -- Network Status
    uint16_t BA_Current_Macrocyle;
    uint16_t BA_Status;
    uint16_t FIELDUAL_Status;

} FIP_NETWORK_STATUS;

#pragma pack()

#endif
